/*
 * Copyright (c) 2007 innoSysTec (R) GmbH, Germany. All rights reserved.
 * Original author: Edmund Wagner
 * Creation date: 05.06.2007
 *
 * Source: $HeadURL$
 * Last changed: $LastChangedDate$
 * 
 * the unrar licence applies to all junrar source and binary distributions 
 * you are not allowed to use this source to re-create the RAR compression algorithm
 * 
 * Here some html entities which can be used for escaping javadoc tags:
 * "&":  "&#038;" or "&amp;"
 * "<":  "&#060;" or "&lt;"
 * ">":  "&#062;" or "&gt;"
 * "@":  "&#064;" 
 */
package de.innosystec.unrar.unpack.ppm;

import de.innosystec.unrar.io.Raw;



/**
 * DOCUMENT ME
 * 
 * @author $LastChangedBy$
 * @version $LastChangedRevision$
 */
public class RarMemBlock extends Pointer
{

	public static final int size = 12;

	private short stamp, NU;

	private int next, prev; // Pointer RarMemBlock

	public RarMemBlock(byte[] mem)
	{
		super(mem);
	}

	public void insertAt(RarMemBlock p)
	{
		RarMemBlock temp = new RarMemBlock(mem);
		prev = p.getAddress();
		temp.setAddress(prev);
		next = temp.getNext();// prev.getNext();
		temp.setNext(this);// prev.setNext(this);
		temp.setAddress(next);
		temp.setPrev(this);// next.setPrev(this);
	}

	public void remove()
	{
		RarMemBlock temp = new RarMemBlock(mem);
		temp.setAddress(prev);
		temp.setNext(next);// prev.setNext(next);
		temp.setAddress(next);
		temp.setPrev(prev);// next.setPrev(prev);
		next = -1;
		prev = -1;
	}

	public int getNext()
	{
		if(mem!=null){
			next = Raw.readIntLittleEndian(mem,  pos+4);
		}
		return next;
	}

	public void setNext(RarMemBlock next)
	{
		setNext(next.getAddress());
	}

	public void setNext(int next)
	{
		this.next = next;
		if (mem != null) {
			Raw.writeIntLittleEndian(mem, pos + 4, next);
		}
	}

	public short getNU()
	{
		if(mem!=null){
			NU = Raw.readShortLittleEndian(mem,  pos+2);
		}
		return NU;
	}

	public void setNU(short nu)
	{
		NU = nu;
		if (mem != null) {
			Raw.writeShortLittleEndian(mem, pos + 2, nu);
		}
	}

	public int getPrev()
	{
		if(mem!=null){
			prev = Raw.readIntLittleEndian(mem,  pos+8);
		}
		return prev;
	}

	public void setPrev(RarMemBlock prev)
	{
		setPrev(prev.getAddress());
	}

	public void setPrev(int prev)
	{
		this.prev = prev;
		if (mem != null) {
			Raw.writeIntLittleEndian(mem, pos + 8, prev);
		}
	}

	public short getStamp()
	{
		if(mem!=null){
			stamp =  Raw.readShortLittleEndian(mem,  pos);
		}
		return stamp;
	}

	public void setStamp(short stamp)
	{
		this.stamp = stamp;
		if (mem != null) {
			Raw.writeShortLittleEndian(mem, pos, stamp);
		}
	}

	@Override
	public boolean setAddress(int pos)
	{
		if (mem == null) {
			return false;
		}
		if (pos < 0 || pos >= mem.length) {
			return false;
		}
		if (pos + size >= mem.length) {
			return false;
		}
		this.pos = pos;

		stamp = Raw.readShortLittleEndian(mem, pos);
		pos += 2;
		NU = Raw.readShortLittleEndian(mem, pos + 2);
		pos += 2;
		next = Raw.readIntLittleEndian(mem, pos);
		pos += 4;
		prev = Raw.readIntLittleEndian(mem, pos);
		return true;
	}

}
