/*
 * Copyright (c) 2007 innoSysTec (R) GmbH, Germany. All rights reserved.
 * Original author: Edmund Wagner
 * Creation date: 01.06.2007
 *
 * Source: $HeadURL$
 * Last changed: $LastChangedDate$
 * 
 * the unrar licence applies to all junrar source and binary distributions 
 * you are not allowed to use this source to re-create the RAR compression algorithm
 * 
 * Here some html entities which can be used for escaping javadoc tags:
 * "&":  "&#038;" or "&amp;"
 * "<":  "&#060;" or "&lt;"
 * ">":  "&#062;" or "&gt;"
 * "@":  "&#064;" 
 */
package de.innosystec.unrar.unpack.ppm;

import de.innosystec.unrar.io.Raw;



/**
 * DOCUMENT ME
 * 
 * @author $LastChangedBy$
 * @version $LastChangedRevision$
 */
public class State extends Pointer
{

	public static final int size = 6;

	private byte symbol;

	private byte freq;

	private int successor; // pointer ppmcontext

	public State(byte[] mem)
	{
		super(mem);
	}

	public byte getFreq()
	{
		if(mem!=null){
			freq = mem[pos+1];
		}
		return freq;
	}

	public void setFreq(byte freq)
	{
		this.freq = freq;
		if (mem != null) {
			mem[pos + 1] = freq;
		}
	}
	public void setValues(State state){
		setFreq(state.getFreq());
		setSuccessor(state.getSuccessor());
		setSymbol(state.getSymbol());
	}

	public State decAddress(){
		setAddress(pos-size);
		return this;
	}
	
	public State incAddress(){
		setAddress(pos+size);
		return this;
	}
	
	public int getSuccessor()
	{
		if(mem!=null){
			successor = Raw.readIntLittleEndian(mem,  pos+2);
		}
		return successor;
	}

	public void setSuccessor(PPMContext successor)
	{
		setSuccessor(successor.getAddress());
	}

	public void setSuccessor(int successor)
	{
		this.successor = successor;
		if (mem != null) {
			Raw.writeIntLittleEndian(mem, pos + 2, successor);
		}
	}

	public byte getSymbol()
	{
		if(mem!=null){
			symbol = mem[pos];
		}
		return symbol;
	}

	public void setSymbol(byte symbol)
	{
		this.symbol = symbol;
		if (mem != null) {
			mem[pos] = symbol;
		}
	}

	@Override
	public boolean setAddress(int pos)
	{
		if (mem == null) {
			return false;
		}
		if (pos < 0 || pos >= mem.length) {
			return false;
		}
		if (pos + size >= mem.length) {
			return false;
		}
		this.pos = pos;
		symbol = mem[pos];
		pos++;
		freq = mem[pos];
		pos++;
		successor = Raw.readIntLittleEndian(mem, pos);
		return true;
	}

}
